from aiohttp import web
import json


__ws_list = None
__subscriptions = {}


def setup_websockets(app):
    global __ws_list
    if not isinstance(__ws_list, list):
        __ws_list = []
        app.router.add_get('/ws', _ws_handler)
        app.on_shutdown.append(_my_shutdown)
    return __ws_list


async def send_to_all(msg: dict):
    for ws in __ws_list:
        await ws.send_str(msg)
    return None


def subscribe(model: str, callback):
    global __subscriptions
    __subscriptions[model] = callback
    return None


async def _ws_handler(request):
    global __subscriptions
    ws = web.WebSocketResponse()
    await ws.prepare(request)

    __ws_list.append(ws)

# Here possiblity to send init-message to Web-Client

    try:
        async for msg in ws:
            if msg.type == web.WSMsgType.TEXT:
                msg = json.loads(msg.data)
                # Callback function from data Model (yet only 'valves')
                func = __subscriptions.get(msg['model'], lambda x: None)
                await func(msg)

    finally:
        __ws_list.remove(ws)
    return ws


async def _my_shutdown(app):
    for ws in __ws_list:
        await ws.close()
    return None



import asyncio
import configargparse
import pkg_resources
import os
import pathlib
# import logging
# import sys

import riego.database
import riego.valves
import riego.parameter
import riego.mqtt
import riego.logger
import riego.timer

from riego.web.routes import setup_routes
from riego.web.middlewares import setup_middlewares
from riego.web.websockets import setup_websockets

from aiohttp import web
import jinja2
import aiohttp_jinja2
import aiohttp_debugtoolbar

from riego.__init__ import __version__


async def on_startup(app):
    app['log'].info("on_startup")
    app['background_timer'] = asyncio.create_task(app['timer'].start_async())
    app['background_mqtt'] = asyncio.create_task(app['mqtt'].start_async())


async def on_cleanup(app):
    app['log'].info("on_cleanup")
    app['background_timer'].cancel()
    await app['background_timer']
    app['background_mqtt'].cancel()
    await app['background_mqtt']


async def on_shutdown(app):
    app['log'].info("on_shutdown")


def main():
    p = configargparse.ArgParser(
        default_config_files=['/etc/riego/conf.d/*.conf', '~/.riego.conf'])
    p.add('-c', '--config', is_config_file=True, env_var='RIEGO_CONF',
          default='riego.conf', help='config file path')
    p.add('-d', '--database', help='Path and name for DB file',
          default='db/riego.db')
    p.add('-e', '--event_log', help='Full path and name for event logfile',
          default='log/event.log')
    p.add('--event_log_max_bytes', help='Maximum Evet Log Size in bytes',
          default=1024*300, type=int)
    p.add('--event_log_backup_count', help='How many files to rotate',
          default=3, type=int)
    p.add('-m', '--mqtt_host', help='IP adress of mqtt host',
          default='127.0.0.1')
    p.add('-p', '--mqtt_port', help='Port of mqtt service',
          default=1883, type=int)
    p.add('--mqtt_client_id', help='Client ID for MQTT-Connection',
          default='riego_controler')
    p.add('--database_migrations_dir',
          help='path to database migrations directory',
          default=pkg_resources.resource_filename('riego', 'migrations'))
    p.add('--base_dir', help='Change only if you know what you are doing',
          default=pathlib.Path(__file__).parent)
    p.add('--http_server_bind_address',
          help='http-server bind address', default='0.0.0.0')
    p.add('--http_server_bind_port', help='http-server bind port',
          default=8080, type=int)
    p.add('--http_server_static_dir',
          help='Serve static html files from this directory',
          default=pkg_resources.resource_filename('riego.web', 'static'))
    p.add('--http_server_template_dir',
          help='Serve template files from this directory',
          default=pkg_resources.resource_filename('riego.web', 'templates'))
    p.add('-v', '--verbose', help='verbose', action='store_true')
    p.add('--version', help='Print version', action='store_true')
    p.add('--enable_aiohttp_debug_toolbar', action='store_true')
    p.add('--enable_asyncio_debug', action='store_true')
    p.add('--enable_timer_dev_mode', action='store_true')

    options = p.parse_args()

    if options.verbose:
        print(p.format_values())

    if options.version:
        print('Version: ', __version__)
        os.sys.exit()

    if options.enable_asyncio_debug:
        asyncio.get_event_loop().set_debug(True)

    if os.name == "posix":
        import uvloop  # pylint: disable=import-error
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

    app = web.Application()

    app['options'] = options
    app['log'] = riego.logger.create_log(options)
    app['event_log'] = riego.logger.create_event_log(options)
    app['db'] = riego.database.Database(app)
    app['mqtt'] = riego.mqtt.Mqtt(app)
    app['valves'] = riego.valves.Valves(app)
    app['parameter'] = riego.parameter.Parameter(app)
    app['timer'] = riego.timer.Timer(app)

    app.on_startup.append(on_startup)
    app.on_cleanup.append(on_cleanup)
    app.on_shutdown.append(on_shutdown)

    aiohttp_jinja2.setup(app, loader=jinja2.FileSystemLoader(
        options.http_server_template_dir))

    setup_routes(app)
    setup_middlewares(app)
    setup_websockets(app)

    if options.enable_aiohttp_debug_toolbar:
        aiohttp_debugtoolbar.setup(app, check_host=False)

    app['log'].info("Start")
    app['event_log'].info('Start')

    web.run_app(app,
                host=options.http_server_bind_address,
                port=options.http_server_bind_port)



"""

"""

from yoyo import step

__depends__ = {'20201210_01_X1eua'}

steps = [
    step(
        """ALTER TABLE options
            RENAME TO parameter;""",
        """ALTER TABLE parameter
            RENAME TO options;"""
    )
]



"""

"""

from yoyo import step

__depends__ = {'20201223_01_mDHp6'}

steps = [
    step(
	"""ALTER TABLE valves ADD COLUMN last_error INTEGER DEFAULT 0""",
	"""ALTER TABLE valves DROP COLUMN last_error"""
	), step(
	"""ALTER TABLE valves ADD COLUMN last_water_amount INTEGER DEFAULT 0""",
	"""ALTER TABLE valves DROP COLUMN last_water_amount"""
	), step(
	"""ALTER TABLE valves ADD COLUMN normal_water_amount INTEGER DEFAULT 0""",
	"""ALTER TABLE valves DROP COLUMN normal_water_amount"""
	), step(
	"""ALTER TABLE valves ADD COLUMN last_duration INTEGER DEFAULT 0""",
	"""ALTER TABLE valves DROP COLUMN last_duration"""
	), step(
	"""ALTER TABLE valves ADD COLUMN last_sheduled timestamp""",
	"""ALTER TABLE valves DROP COLUMN last_sheduled"""
	), 

]



import asyncio

from gmqtt import Client as MQTTClient
from gmqtt.mqtt.constants import MQTTv311


class Mqtt:
    def __init__(self, app):
        self.__options = app['options']
        self.__log = app['log']
        self.client = MQTTClient(self.__options.mqtt_client_id)
        self.client.on_connect = self._on_connect
        self.client.on_message = self._on_message
        self.client.on_disconnect = self._on_disconnect
        self.client.on_subscribe = self._on_subscribe

    def _on_connect(self, client, flags, rc, properties):
        self.__log.debug('MQTT Connected')
        self.client.subscribe('TEST/#', qos=0)

    def _on_message(self, client, topic, payload, qos, properties):
        self.__log.debug('MQTT RECV MSG:', payload)

    def _on_disconnect(self, client, packet, exc=None):
        self.__log.debug('MQTT Disconnected')

    def _on_subscribe(self, client, mid, qos, properties):
        self.__log.debug('MQTT SUBSCRIBED')

    async def start_async(self):
        await self.client.connect(self.__options.mqtt_host,
                                  port=self.__options.mqtt_port,
                                  keepalive=10,
                                  version=MQTTv311)
        while True:
            try:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                self.__log.debug('MQTT: trapped cancel')
                break
        await self.client.disconnect()



import paho.mqtt.client
import asyncio


class Mqtt:
    def __init__(self, app):
        options = app['options']
        self.__log = app['log']
        self.client = paho.mqtt.client.Client()
        self.client.on_connect = self._on_connect
        self.client.on_disconnect = self._on_disconnect
        self.client.on_message = self._on_message
        self.client.connect(options.mqtt_host, options.mqtt_port, 10)
        self.__wait_for_first_conn()

    def __wait_for_first_conn(self):
        self.client.loop()
        self.client.loop()

    def _on_connect(self, client, userdata, flags, rc):
        if rc == 0:
            self.client.subscribe("cmnd/controler/#")
            self.__log.info("Mqtt Connection established")
        else:
            self.__log.critical("Bad mqtt connection, rc: " + str(rc))

    # The callback for when a PUBLISH message is received from the server.

    def _on_message(self, client, userdata, msg):
        self.__log.debug("catch all " + msg.topic+" "+str(msg.payload))

    def _on_disconnect(self, client, userdata, rc):
        if rc != 0:
            self.__log.error(
                "Unexpected mqtt disconnection rc: " + str(rc))

    async def start_async(self):
        while True:
            try:
                # TODO: convert to Task
                self.client.loop()
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                self.__log.debug("mqtt: trapped cancel")
                break
        self.client.disconnect()



import sqlite3
from pathlib import Path
from yoyo import read_migrations
from yoyo import get_backend


class Database:
    def __init__(self, app):
        options = app['options']
        log = app['log']
        self.conn = None
        Path(options.database).parent.mkdir(parents=True, exist_ok=True)
        backend = get_backend('sqlite:///' + options.database)
        migrations = read_migrations(options.database_migrations_dir)
        with backend.lock():
            backend.apply_migrations(backend.to_apply(migrations))
        try:
            self.conn = sqlite3.connect(options.database,
                                        detect_types=sqlite3.PARSE_DECLTYPES |
                                        sqlite3.PARSE_COLNAMES)
            self.conn.row_factory = sqlite3.Row
        except sqlite3.Error as error:
            log.info("sqlite3.connect error: " + str(error))
            if (self.conn):
                self.conn.close()

    def __del__(self):
        try:
            self.conn.close()
        except:
            print('Error')



"""

"""

from yoyo import step

__depends__ = {'20201209_03_YHuBM'}

steps = [
    step(
	"""CREATE TABLE boxes
            (id  	INTEGER PRIMARY KEY AUTOINCREMENT,
            name	TEXT UNIQUE,
			remark	TEXT,
            first_seen	timestamp DEFAULT CURRENT_TIMESTAMP,
            has_config	INTEGER DEFAULT 0,
			is_online	INTEGER)""",
	"""DROP TABLE boxes"""
	)
	
	
]





class Parameter:
    def __init__(self, app):
        self.db_conn = app['db'].conn
        self.mqtt = app['mqtt']

    def get(self, key):
        c = self.db_conn.cursor()
        c.execute('SELECT * from parameter WHERE key=?',
                  (key,))
        ret = c.fetchone()
        c.close()
        if ret is not None:
            ret = (ret['value'])
        return ret

    def update(self, key, value):
        c = self.db_conn.cursor()
        c.execute('UPDATE parameter SET value=? WHERE key=?',
                  (value, key))
        self.db_conn.commit()
        c.close()

    def insert(self, key, value):
        c = self.db_conn.cursor()
        c.execute('INSERT INTO parameter (key,value) VALUES (?,?)',
                  (value, key))
        self.db_conn.commit()
        c.close()

    def delete(self, key):
        c = self.db_conn.cursor()
        c.execute('DELETE FROM parameter WHERE key=?',
                  (key,))
        self.db_conn.commit()
        c.close()



"""

"""

from yoyo import step

__depends__ = {}

steps = [
    step(
	"""CREATE TABLE valves
            (id  	INTEGER PRIMARY KEY AUTOINCREMENT,
            name	TEXT UNIQUE,
            topic	TEXT UNIQUE,
            duration	INTEGER DEFAULT 0,
            interval	INTEGER DEFAULT 0,
            last_run	timestamp DEFAULT CURRENT_TIMESTAMP,
            is_running	INTEGER DEFAULT 0,
            is_enabled	INTEGER DEFAULT 1)""",
	"""DROP TABLE valves"""
	),
	step(
	"""CREATE TABLE options
            (id	    INTEGER PRIMARY KEY AUTOINCREMENT,
            key	    TEXT UNIQUE,
            value   TEXT)""",
	"""DROP TABLE options"""
	)
]



"""

"""

from yoyo import step

__depends__ = {'20201209_02_i5rFY'}

steps = [
    step(
	"INSERT INTO options (key,value) VALUES ('startTime', '19:00')",
	"DELETE FROM options WHERE key = 'startTime'"
	),
    step(
	"INSERT INTO options (key,value) VALUES ('maxDuration', '4')",
	"DELETE FROM options WHERE key = 'maxDuration'"
	),
    step(
	"INSERT INTO options (key,value) VALUES ('heat', '0')",
	"DELETE FROM options WHERE key = 'heat'"
	),
    step(
	"INSERT INTO options (key,value) VALUES ('rain', '0')",
	"DELETE FROM options WHERE key = 'rain'"
	),
    step(
	"INSERT INTO options (key,value) VALUES ('pause', '0')",
	"DELETE FROM options WHERE key = 'pause'"
	),
    step(
	"INSERT INTO options (key,value) VALUES ('disable', '0')",
	"DELETE FROM options WHERE key = 'enable'"
	)

]



from datetime import datetime, timedelta
import asyncio


class Timer():
    def __init__(self, app):
        self.__log = app['log']
        self.__valves = app['valves']
        self.__parameter = app['parameter']
        self.__options = app['options']

    async def start_async(self):
        v = self.__valves.get_next()
        while True:
            try:
                if v.is_running:
                    td = timedelta(minutes=v.duration)
                    if self.__options.enable_timer_dev_mode:
                        td = timedelta(minutes=0, seconds=v.duration)

                    if datetime.now() - v.last_run > td:
                        # Laufzeit erreicht
                        await v.set_is_running(0)
                        self.__log.debug("valveOff " + v.name)
                    else:
                        pass
                else:
                    td = timedelta(days=v.interval)
                    if self.__options.enable_timer_dev_mode:
                        td = timedelta(days=0, seconds=v.interval)

                    if (datetime.now() - v.last_run > td and v.is_enabled and
                            datetime.now().strftime('%H:%M') >=
                            self.__parameter.get('startTime')):
                        # Intervall erreicht
                        await v.set_is_running(1)
                        self.__log.debug("valveOn " + v.name)
                    else:
                        v = self.__valves.get_next()
                        self.__log.debug("NextValve: " + v.name)
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                self.__log.debug("Timer: trapped cancel")
                break
        await v.set_is_running(0)



from datetime import datetime
import json
import riego.web.websockets


class Valves():
    def __init__(self, app):
        db_conn = app['db'].conn
        mqtt = app['mqtt']
        event_log = app['event_log']
        self.log = app['log']
        self._valves = []
        self.idx_valves = -1
        for row in db_conn.execute('select * from valves'):
            v = Valve(db_conn, mqtt, event_log, row)
            self._valves.append(v)

        riego.web.websockets.subscribe('valves', self._ws_handler)

    def get_next(self):
        self.idx_valves += 1
        if len(self._valves) == 0:
            return None
        if self.idx_valves >= len(self._valves):
            self.idx_valves = 0
        return self._valves[self.idx_valves]

    def get_dict(self):
        ret = {}
        for v in self._valves:
            ret[v.id] = v.get_dict()
        return ret

    def get_valve_by_id(self, id: int):
        for v in self._valves:
            if v.id == int(id):
                return v
        return None

    async def _ws_handler(self, msg: dict):
        """
        Call function from Object Valves according to msg['propo']
        """
        if not msg['action'] == 'update':
            return None
        valve = self.get_valve_by_id(msg['id'])
        func = getattr(valve, "set_" + msg['prop'])
        await func(msg['value'])


class Valve():

    def __init__(self, db_conn, mqtt, event_log, row):
        self.__db_conn = db_conn
        self.__mqtt = mqtt
        self.__event_log = event_log

        self.bool_to_int = {'true': 1, 'false': 0, True: 1, False: 0,
                            'True': 1, 'False': 0, 'on': 1, 'off': 0,
                            'On': 1, 'Off': 0, 'ON': 1, 'OFF': 0}

        self.__id = row['id']
        self.__name = row['name']
        self.__topic = row['topic']
        self.__duration = row['duration']
        self.__interval = row['interval']
        self.__last_run = row['last_run']
        self.__is_running = row['is_running']
        self.__is_enabled = row['is_enabled']
        self.__remark = row['remark']

    @ property
    def id(self):
        return self.__id

    @ property
    def name(self):
        return self.__name

    async def set_name(self, val: str):
        with self.__db_conn:
            self.__db_conn.execute(
                'UPDATE valves SET name = ?  WHERE id = ?',
                (val, self.__id))
        await self.send_status_with_websocket('name', val)
        self.__name = val

    @ property
    def topic(self):
        return self.__topic

    async def set_topic(self, val: str):
        with self.__db_conn:
            self.__db_conn.execute(
                'UPDATE valves SET topic = ?  WHERE id = ?',
                (val, self.__id))
        await self.send_status_with_websocket('topic', val)
        self.__topic = val

    @ property
    def duration(self):
        return self.__duration

    async def set_duration(self, val: int):
        val = int(val)
        with self.__db_conn:
            self.__db_conn.execute(
                'UPDATE valves SET duration = ?  WHERE id = ?',
                (val, self.__id))
        await self.send_status_with_websocket('duration', val)
        self.__duration = val

    @ property
    def interval(self):
        return self.__interval

    async def set_interval(self, val: int):
        val = int(val)
        with self.__db_conn:
            self.__db_conn.execute(
                'UPDATE valves SET interval = ?  WHERE id = ?',
                (val, self.__id))
        await self.send_status_with_websocket('interval', val)
        self.__interval = val

    @ property
    def last_run(self):
        return self.__last_run

    def set_last_run(self):
        raise NotImplementedError

    @ property
    def is_running(self):
        return self.__is_running

    async def set_is_running(self, val):
        val = self.bool_to_int[val]
        with self.__db_conn:
            self.__db_conn.execute(
                'UPDATE valves SET is_running = ?  WHERE id = ?',
                (val, self.__id))
        self.__mqtt.client.publish(self.__topic, val)
        self.__event_log.info(str(val) + ';' + self.name)
        await self.send_status_with_websocket('is_running', val)
        if val == 1:
            self.__last_run = datetime.now()
            with self.__db_conn:
                self.__db_conn.execute(
                    'UPDATE valves SET last_run = ?  WHERE id = ?',
                    (self.__last_run, self.__id))
            await self.send_status_with_websocket('last_run',
                                                  str(self.__last_run))
        self.__is_running = val

    @ property
    def is_enabled(self):
        return self.__is_enabled

    async def set_is_enabled(self, val):
        val = self.bool_to_int[val]

        with self.__db_conn:
            self.__db_conn.execute(
                'UPDATE valves SET is_enabled = ?  WHERE id = ?',
                (val, self.__id))
        await self.send_status_with_websocket('is_enabled', val)
        self.__is_enabled = val

    @ property
    def remark(self):
        return self.__remark

    async def set_remark(self, val):
        with self.__db_conn:
            self.__db_conn.execute(
                'UPDATE valves SET remark = ?  WHERE id = ?',
                (val, self.__id))
        await self.send_status_with_websocket('remark', val)
        self.__remark = val

    def get_dict(self):
        ret = {'id': self.__id,
               'name': self.__name,
               'topic': self.__topic,
               'duration': self.__duration,
               'interval': self.__interval,
               'last_run': str(self.__last_run),
               'is_running': self.__is_running,
               'is_enabled': self.__is_enabled,
               'remark': self.__remark}
        return ret

    async def send_status_with_websocket(self, prop, value):
        ret = {
            'action': "status",
            'model': "valves",
            'id': self.__id,
            'prop': prop,
            'value': value,
        }
        ret = json.dumps(ret)
        await riego.web.websockets.send_to_all(ret)
        return ret
